/* ser_sr24.c
 * 
 * Provides the functions and data structures for handling
 * the model SR24 device.
 * 
 * Copyright (C) 2000, Winford Engineering.
 * See file LICENSE.txt for details.
 * 
 * Winford Engineering
 * www.winfordeng.com
 */
#ifdef WIN95
#define COMMONWIN
#endif
#ifdef WINNT
#define COMMONWIN
#endif


#include "ser_sr24.h"
#include "ser_sr24_helper.h"
#include "ser_threads.h"
#include "ser_serial.h"
#include "ser_error.h"

#ifdef COMMONWIN
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <malloc.h>

#define CALLCONVENTION __stdcall
#endif

#ifdef LINUX
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <malloc.h>

#define CALLCONVENTION
#endif



ser_sr24_data CALLCONVENTION *ser_sr24_open(char* serial, unsigned int baud, int *result)
{/* returns a pointer to the structure if the device was successfully 
  * opened and initialized.  On failure, a NULL pointer is returned.
  * 
  * *result provides more in-depth result codes.  result may be NULL
  * if no more information is desired.  On success, and if result is not 
  * null, *result will be zero.  On failure, *result will contain a
  * negative error code, as listed in the header file.
  */
	int retval;
	int erresult=0; /* error code to be returned */
	ser_sr24_data *data;

	if( (data=(ser_sr24_data*)malloc(sizeof(ser_sr24_data))) == 0 )
	{
		if(result)
			*result = SER_MALLOC_ERROR; 
		return(0);
	}
	
	if( ser_mutex_init(&(data->mutex)) )
	{	erresult = SER_THREAD_ERROR;
		goto error_1;
	}

	if( ser_event_init(&(data->event)) )
	{
		erresult = SER_THREAD_ERROR;
		goto error_2;
	}

	if( ser_event_init(&(data->synch_event)) )
	{
		erresult = SER_THREAD_ERROR;
		goto error_3;
	}

	/* open serial port */
	if( (retval=ser_serial_open(&(data->serial), serial, baud)) )
	{
		erresult=retval; /* return the error if unsuccessful */
		goto error_4;
	}
	/* set timeout so reading thread will awake periodically */
	else if( (retval=ser_serial_timeout(&(data->serial), 1500)) )
		erresult=retval;
	/* issue command to reset device */
	else if( ser_serial_write(&(data->serial), "R\r", 2)<0 )
		erresult=SER_SERIAL_ERROR;
	else
	{
		/* reset all the status data to reflect the reset state of the device */
		ser_sr24_reset_board_data(data);

		/* initialize internal data members */
		data->quitflag = 0;
		data->event_quitflag = 0;
		data->portclean = 0;
		data->bitclean[0] = 0;
		data->bitclean[1] = 0;
		data->bitclean[2] = 0;
		data->analogclean = 0;
		data->timeout = 1000; /* default to 1 second timeout for user functions */
		data->event_handler.type = 0;
		data->last_event = 0;

		if( ser_thread_create(ser_sr24_readthread, (void*)data) )
			erresult=SER_READTHREAD_ERROR;
	}

	if(erresult)/* if an error has occurred */
	{         /* NOTE: since creating the thread is the last operation, and
	           * it is not performed unless everything else succeeds, if the
	           * thread creationg fails, then we can safely return an error
	           * without having to cancel the thread.
	           */
		goto error_5;
	}

	if( ser_sr24_verify_connection(data) )
	{ /* if the verify_connection function fails, we'll stop everything
	   * and return an error.  The user will not end up getting a valid
	   * file handle, but rather an error that we return.
	   */
		ser_sr24_close(data); /* close will also free the structure we allocated */
		if(result)
			*result = SER_DEVICE_NOTFOUND;
		return(0);
	}

	if(result)
		*result = 0;

#ifdef COMMONWIN
	/* If we're on a Windows machine, keep track of this opened device */
	ser_sr24_table_add(data);
#endif

	return(data); /* return a pointer to the structure we just allocated and initialized */

error_5: /* need to close serial connection */
	ser_serial_close(&(data->serial));
	
error_4: /* need to destroy both events and mutex. */
	ser_event_destroy(&(data->synch_event));
	
error_3: /* failed to init synch_event, destroy event */
	ser_event_destroy(&(data->event));

error_2: /* failed to init event, destroy mutex */
	ser_mutex_destroy(&(data->mutex));

error_1: /* just free the memory and return an error */
	free(data);

	if(result)
		*result = erresult;
	return(0);
}


int CALLCONVENTION ser_sr24_verify_connection(ser_sr24_data* data)
{
/* Verifies that we have a good connection to the device.
 * Returns 0 if connection is good.
 * Returns SER_TIMEOUT if no response was received from the device
 */
	unsigned int tout;
	unsigned int * ptout;
	int retval;
	int result;

	if(data == 0)
		return(SER_INVALID_HANDLE);
	
	tout=data->timeout;
	if(tout)
		ptout = &tout;
	else
		ptout = 0;

	/* lock the mutex so we can modify the verifyflag */
	if( ser_event_prewait(&(data->event)) )
		return(SER_THREAD_ERROR);

	/* we'll send out a ping command to verify the connection
	 */
	data->verifyflag=1; /* set flag then we'll know when the response was received */

	if( ser_serial_write(&(data->serial), "P\r", 2)<0 )
	{/* if error, bail out */
		ser_event_release(&(data->event));
		return(SER_SERIAL_ERROR);
	}

	retval=0;
	while( data->verifyflag ) /* wait until the thread sets verifyflag 
	                         * back to zero */
	{
		if( (retval=ser_event_wait(&(data->event), ptout)) )
			/* if error or timeout */
			break;
	}

	result = data->verifyflag;

	ser_event_release(&(data->event));

	if(result == 0)	
		return(0); /* indicate good connection */
	else if(retval == 1)
		return(SER_TIMEOUT);
	else
		return(SER_THREAD_ERROR);
}

int CALLCONVENTION ser_sr24_reset_board(ser_sr24_data *data)
{
	/* This function resets the serial board, and also updates the internal
	 * state of the API to match the reset state of the board.
	 * 
	 * This function does NOT affect internal items like timeout value,
	 * or event handler specifications.
	 * 
	 * This function can time out, but it is quite unlikely that it 
	 * ever would.
	 */
	unsigned int tout;
	unsigned int *ptout;
	int retval;

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;

	if(tout)
		ptout = &tout;
	else
		ptout = 0;
		
	if( (retval=ser_mutex_wait(&(data->mutex), ptout)) )
	{
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	/* reset our own data structure to match what the board 
	 * will be reset to
	 */
	ser_sr24_reset_board_data(data);
	
	ser_mutex_release(&(data->mutex));

	if( ser_serial_write(&(data->serial), "R\r", 2)<0 )
		return(SER_SERIAL_ERROR);
		
	return(0); /* success */
}

int CALLCONVENTION ser_sr24_get_portmode(ser_sr24_data *data, int port, int property)
{
/* Checks the configuration of the specified port.  
 * Returns the state of the property specified.
 */
	if(data == 0)
		return(SER_INVALID_HANDLE);
	
	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	switch(property)
	{
	case SR24_BUFFER:
	case SR24_DIRECTION:
	case SR24_TYPE:
		return( ((int)pow(2, property) & data->mode[port-1]) != 0 );
		break;
	}

	return(SER_INVALID_PROPERTY);
}

int CALLCONVENTION ser_sr24_set_portmode(ser_sr24_data* data, int port, int property, int mode)
{
/* configures and checks the configuration for a port, based
 * on the value the user passes for mode
 */
	char buf[6]; /* currently, longest cmd is 4 + LF + NULL */
	unsigned int tout;
	unsigned int *ptout;
	int retval;
	int erresult=0;
	int fail=0; /* if nonzero at end, an error or something occurred. return erresult */

	if(data == 0)
		return(SER_INVALID_HANDLE);
	
	
	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	tout=data->timeout;

	if(tout)
		ptout = &tout;
	else
		ptout = 0;
		
	if( (retval=ser_mutex_wait(&(data->mutex), ptout)) )
	{
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	switch(property)
	{
	case SR24_BUFFER:
		{
			switch(mode)
			{
			case SR24_BUFFER_ENABLED:
				data->mode[port-1] &= ~(int)pow(2, SR24_BUFFER);
				sprintf(buf, "E%d\r", port);
				break;
			case SR24_BUFFER_DISABLED:
				data->mode[port-1] |= (int)pow(2, SR24_BUFFER);
				sprintf(buf, "X%d\r", port);
				break;
			default:
				erresult=SER_INVALID_PORTMODE;
				fail=1;
			}
		}
		break;
	case SR24_DIRECTION:
		{
			switch(mode)
			{
			case SR24_DIRECTION_INPUT:
				data->mode[port-1] &= ~(int)pow(2, SR24_DIRECTION);
				sprintf(buf, "S%d,0\r", port);/* string to set to input state */
				break;
			case SR24_DIRECTION_OUTPUT:
				data->mode[port-1] |= (int)pow(2, SR24_DIRECTION);
				sprintf(buf, "S%d,1\r", port);/* string to set to output state */
				break;
			default:
				erresult=SER_INVALID_PORTMODE;
				fail=1;
			}
		}
		break;
	case SR24_TYPE:
		{
			switch(mode)
			{
			case SR24_TYPE_DIGITAL:
				if(port != 1) /* ports other than 1 are always digital on this device */
				{
					erresult=0; /* no error */
					fail=1; /* don't need to write anything to board */
				}
					
				data->mode[port-1] &= ~(int)pow(2, SR24_TYPE);
				sprintf(buf, "D%d\r", port); /* string to set port to digital */
				break;
			case SR24_TYPE_ANALOG:
				if(port != 1) /* port 1 is the only analog-capable port on this device */
				{	
					erresult=SER_INVALID_PORTMODE;
					fail=1;
				}

				/* Serial board clears all bit and byte events when putting port into
				 * analog mode.  We should do the same.
				 */
				 		
				memset(data->bitevent[0], 0, sizeof(data->bitevent[0]));
				data->portevent[0] = 0;

				data->mode[port-1] |= (int)pow(2, SR24_TYPE);
				sprintf(buf, "A%d\r", port);
				break;
			default:
				erresult=SER_INVALID_PORTMODE;
				fail=1;
			}
		}
		break;
	default:
		erresult=SER_INVALID_PROPERTY;
		fail=1;
	}

	ser_mutex_release(&(data->mutex));

	if(fail)
		return(erresult);
	
	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
		return(SER_SERIAL_ERROR);
	return(0); /* success */
}

int CALLCONVENTION ser_sr24_set_timeout(ser_sr24_data* data, unsigned int timeout)
{
	/* Sets the timeout value for all other functions that may timeout,
	 * and that aren't passed an explicit timeout value.
	 * 
	 * timeout is specified in milliseconds.  If timeout is zero,
	 * the functions will never time out.
	 */
	if(data == 0)
		return(SER_INVALID_HANDLE);

	 
	if( ser_mutex_wait(&(data->mutex), 0) )
		return(SER_THREAD_ERROR);

	data->timeout = timeout;

	if( ser_mutex_release(&(data->mutex)) )
		return(SER_THREAD_ERROR);
	
	return(0);
}

int CALLCONVENTION ser_sr24_input_byte(ser_sr24_data* data, int port, int* value)
{
/* this function returns the value of a port.  This function may time out,
 * based on the setting from ser_sr24_set_timeout()
 */
	unsigned int tout;
	unsigned int *ptout;
	int retval;
	int invalue; /* value that came in from serial line */
	char buf[4]; /* space for max out string (2 characters) plus return and null */

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;
	if(tout)
		ptout = &tout;
	else
		ptout = 0;
	
	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	/* NOTE: We won't check for direction or enabled in order to give
	 * the user full flexibility
	 */

	/* lock the mutex so we can modify the clean bitmask */
	if( ser_event_prewait(&(data->event)) )
		return(SER_THREAD_ERROR);

	
	/* set the port we want to be dirty */
	data->portclean &= ~(int)(pow(2, port-1)); //set the bit to zero.

	/* construct the output string */
		/* this should never overrun the buffer because of the checks
		 * on port above
		 */	
	sprintf(buf, "I%d\r", port);
	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
	{/* if error, bail out */
		ser_event_release(&(data->event));
		return(SER_SERIAL_ERROR);
	}

	retval=0;
	while( (data->portclean & (int)pow(2, port-1)) == 0 )
	{
		if( (retval=ser_event_wait(&(data->event), ptout)) )
			/* if error or timeout */
			break;
	}
	if(retval)
	{
		ser_event_release(&(data->event));
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	/* IF we are here, then data has come in, we still
	 * own the event, and the clean bit is set.
	 */

	invalue = data->portvalue[port-1];

	/* set the port to be dirty again*/
	data->portclean &= ~(int)(pow(2, port-1)); //set the bit to zero.
	
	ser_event_release(&(data->event));

	*value = invalue;
	return(0);
}

int CALLCONVENTION ser_sr24_output_byte(ser_sr24_data* data, int port, int value)
{
	char buf[8]; /*space for maximum output string (6 characters) plus return and null */

	if(data == 0)
		return(SER_INVALID_HANDLE);
	
	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	if( (value < 0) || (value > 255) )
	/* invalid data */
		return(SER_INVALID_PORTVALUE);

	/* We won't check for enabled or direction in order
	 * to give the user full flexibility
	 */

	/* construct the output string (in hexadecimal)*/
		/* this should never overrun the buffer because of the checks
		 * on port and value above
		 */	
	sprintf(buf, "O%d=%x\r", port, value);

	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
		return(SER_SERIAL_ERROR);
	return(0); /* success */
}

int CALLCONVENTION ser_sr24_input_bit(ser_sr24_data *data, int port, int bit, int *value)
{
	/* Reads an individual bit on an output port 
	 * No checks are made on the state of the port (output mode or being enabled)
	 * in order to give the programmer full flexibility.
	 * For example, if the programmer wants to put a value on the port, and then
	 * later enable the buffer, this is possible because of no checks.
	 * 
	 * Checks are made on the port number (1-3) and the bit number (0-7).  
	 * They must be in the range.
	 * 
	 * If the bit is on, 1 will be written to *value, otherwise 0
	 */
	unsigned int tout;
	unsigned int *ptout;
	int retval;
	int invalue; /* value that came in from serial line */
	char buf[6]; /*space for maximum output string (4 characters) plus return and null */

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;
	if(tout)
		ptout = &tout;
	else
		ptout = 0;

	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	if( (bit < 0) || (bit > SER_SR24_BIT_MAX) )
	/* invalid bit number */
		return(SER_INVALID_BIT);

	/* We won't check for enabled or direction in order
	 * to give the user full flexibility
	 */

	/* lock the event because we're about to modify the clean flags */
	if( ser_event_prewait(&(data->event)) )
		return(SER_THREAD_ERROR);

	/* set the bit we want to be dirty */
	data->bitclean[port-1] &= ~(int)(pow(2, bit)); //set the bit to zero.

	/* construct the output string */
		/* this should never overrun the buffer because of the checks
		 * on port above
		 */	
	sprintf(buf, "i%d.%d\r", port, bit);
	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
	{/* if error, bail out */
		ser_event_release(&(data->event));
		return(SER_SERIAL_ERROR);
	}

	/* we still own the event.  It'll be released during the waiting period */
	retval=0;
	while( (data->bitclean[port-1] & (int)pow(2, bit)) == 0 )
	{
		if( (retval=ser_event_wait(&(data->event), ptout)) )
			/* if error or timeout */
			break;
	}
	if(retval)
	{
		ser_event_release(&(data->event));
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	/* IF we are here, then data has come in, we still
	 * own the event, and the clean bit is set.
	 */

	invalue = ( data->bitvalue[port-1] & (int)pow(2, bit) ) ? 1 : 0;
	
	/* set the port to be dirty again*/
	data->bitclean[port-1] &= ~(int)(pow(2, bit)); //set the bit to zero.
	
	ser_event_release(&(data->event));


	*value = invalue;
	return(0);

}

int CALLCONVENTION ser_sr24_output_bit(ser_sr24_data *data, int port, int bit, int value)
{
	/* Sets an individual bit on an output port 
	 * No checks are made on the state of the port (output mode or being enabled)
	 * in order to give the programmer full flexibility.
	 * For example, if the programmer wants to put a value on the port, and then
	 * later enable the buffer, this is possible because of no checks.
	 * 
	 * Checks are made on the port number (1-3) and the bit number (0-7).  
	 * They must be in the range.
	 * 
	 * Any nonzero value for value will turn the bit on.  Zero turns it off.
	 */
	char buf[8]; /*space for maximum output string (6 characters) plus return and null */

	if(data == 0)
		return(SER_INVALID_HANDLE);

	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	if( (bit < 0) || (bit > SER_SR24_BIT_MAX) )
	/* invalid bit number */
		return(SER_INVALID_BIT);

	/* We won't check for enabled or direction in order
	 * to give the user full flexibility
	 */

	/* construct the output string */
		/* this should never overrun the buffer because of the checks
		 * on port and bit above
		 */	
	sprintf(buf, "o%d.%d=%d\r", port, bit, (value) ? 1 : 0);

	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
		return(SER_SERIAL_ERROR);
	return(0); /* success */
}

int CALLCONVENTION ser_sr24_input_analog(ser_sr24_data *data, int channel, int *value)
{
	/* This function reads the value of an analog channel.
	 * 
	 * Port 1 is the only port that can be put into analog mode,
	 * so this function does not need a port parameter.
	 * 
	 * Also, an analog reading when a port is in digital mode is
	 * meaningless, so this function will return an error if the
	 * port is in digital mode.
	 */

	unsigned int tout;
	unsigned int *ptout;
	int retval;
	int invalue; /* value that came in from serial line */
	char buf[5]; /*space for maximum output string (3 characters) plus return and null */

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;
	if(tout)
		ptout = &tout;
	else
		ptout = 0;

	if( (channel < SER_SR24_CHAN_MIN) || (channel > SER_SR24_CHAN_MAX) )
	/* bad channel specified */
		return(SER_INVALID_CHANNEL);

	/* In this case, we do check to make sure that the port is
	 * in analog mode before we proceed
	 */

	if( ((int)pow(2, SR24_TYPE) & data->mode[0]) == 0 )
	{/* then the port is in digital mode. Error. */
		return(SER_WRONG_PORTMODE);
	}
	
	/* lock the event because we're about to modify the clean flags */
	if( ser_event_prewait(&(data->event)) )
		return(SER_THREAD_ERROR);

	/* set the channel we want to be dirty */
	data->analogclean &= ~(int)(pow(2, channel)); //set the bit to zero.

	/* construct the output string */
		/* this should never overrun the buffer because of the checks
		 * on channel above
		 */	
	sprintf(buf, "a.%d\r", channel);
	if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
	{/* if error, bail out */
		ser_event_release(&(data->event));
		return(SER_SERIAL_ERROR);
	}

	/* we still own the event.  It'll be released during the waiting period */
	retval=0;
	while( (data->analogclean & (int)pow(2, channel)) == 0 )
	{
		if( (retval=ser_event_wait(&(data->event), ptout)) )
			/* if error or timeout */
			break;
	}
	if(retval)
	{
		ser_event_release(&(data->event));
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	/* IF we are here, then data has come in, we still
	 * own the event, and the clean bit is set.
	 */

	invalue = data->analogvalue[channel];
	
	/* set the channel to be dirty again*/
	data->analogclean &= ~(int)(pow(2, channel)); //set the bit to zero.
	
	ser_event_release(&(data->event));


	*value = invalue;
	return(0);
}

int CALLCONVENTION ser_sr24_register_event(ser_sr24_data* data, int port, int bit, int id)
{
/*registers an event, sets the id, and sends out the command
 *to the device to enable the event.
 *
 *If bit is -1, this registers a port event,
 *otherwise if the bit is within the valid range,
 *it registers a bit event.
 *
 *id may take on any value.  A value of 0 indicates
 *that the event should be disabled. All other
 *values indicate the event should be enabled.
 *
 *NOTE: Registering a port event automatically
 *      disables all bit events on that port.
 *      
 *  
 */
	char buf[6];
	unsigned int tout;
	unsigned int * ptout;
	int retval;
	int result=0;

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;
	if(tout)
		ptout = &tout;
	else
		ptout = 0;
		
	if( (port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX) )
	/* bad port specified */
		return(SER_INVALID_PORT);

	if( (bit < -1) || (bit > SER_SR24_BIT_MAX) )
	/* allow -1 to specify port event */
		return(SER_INVALID_BIT);

	if( id < 0 ) /* out of valid range */
		return(SER_INVALID_EVENTID);
		
	if( (retval=ser_mutex_wait(&(data->mutex), ptout)) )
	{
		if(retval == 1)
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	if( (port == 1) && (data->mode[0] & (int)pow(2, SR24_TYPE)) ) /* no events can be registered on port 1 if it's in analog mode */
	{
		result = SER_WRONG_PORTMODE;
	}
	else if(bit == -1)
	{/* port event */
		data->portevent[port-1]=id;

		
		if(id)
		{/*enable event*/

			/* Since a port event is being registered, all bit events on that port should be cleared.
			 * Note that the serial board only clears bit events when a byte event is
			 * turned on, not when one is turned off.
			 */

			memset(data->bitevent[port-1], 0, sizeof(data->bitevent[0]));

			sprintf(buf, "V%d\r", port);
			if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
				result=SER_SERIAL_ERROR;
		}
		else
		{/*disable event*/
			sprintf(buf, "C%d\r", port);
			if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
				result=SER_SERIAL_ERROR;
		}
	}
	else
	{/* bit event */
		if(data->portevent[port-1]) /* if the event is enabled on the port, that's an error */
		{
			result = SER_EVENT_CONFLICT;
		}

		data->bitevent[port-1][bit]=id;

		if(id)
		{/*enable event*/
			sprintf(buf, "v%d.%d\r", port, bit);
			if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
				result=SER_SERIAL_ERROR;
		}
		else
		{/*disable event*/
			sprintf(buf, "c%d.%d\r", port, bit);
			if( ser_serial_write(&(data->serial), buf, strlen(buf))<0 )
				result=SER_SERIAL_ERROR;
		}
	}

	ser_mutex_release(&(data->mutex));
	return(result);
}

int CALLCONVENTION ser_sr24_set_event_handler(ser_sr24_data* data, ser_event_spec* event)
{
	/* The application calls this function to tell the API how it wants
	 * to be notified of serial board events.  There are a few possibilities
	 * including no notification.
	 * 
	 * For convenience, if a null pointer is passed for event, event notification
	 * will be turned off, just as if a structure was passed with type of zero.
	 * 
	 * Return values:
	 * 0 for success
	 * negative error code on failure.
	 * 
	 * If the function fails, the API will use no event notification type.
	 * For example, if initially the API is using callbacks as a notification type,
	 * and the application calls this function and specifies an invalid notification
	 * type, causing the function to fail, the the API will not return to the callbacks
	 * but will instead end up having no notification.  If desired, the application could
	 * re-call this function to set notification back to callbacks
	 */
	
	unsigned int tout;
	unsigned int *ptout;
	ser_event_spec evspec; /* only for use if null pointer is passed in */
	int retval;
	int result=0;/* the value we'll return */

	if(data == 0)
		return(SER_INVALID_HANDLE);

	tout=data->timeout;
	if(tout) /* if timeout value is 0, set the pointer to be 0 so
	          * the wait functions will not time out */
		ptout = &tout;
	else
		ptout = 0;

	if(event == 0) /* if a null pointer was passed, make our own event structure */
	{
		evspec.type = 0;
		event = &evspec;
	}

	/* do a quick check to make sure the same type isn't being requested as
	 * already exists.  If it's the same, there's no reason to destroy it
	 * just to recreate it.
	 */
	if(data->event_handler.type == event->type)
		return(0); /* Don't need to do anything. return success */
	
	if( (retval=ser_mutex_wait(&(data->mutex), ptout)) )
	{
		if(retval == 1)/*timeout ocurred*/
			return(SER_TIMEOUT);
		else
			return(SER_THREAD_ERROR);
	}

	switch(data->event_handler.type) /* depending on the previous event handler, we may need to do cleanup */
	{
	case 1: /* need to get rid of the event thread */
		ser_sr24_close_eventthread(data, 0);
		break;
	/* no cleanup necessary for plain Windows messages */
#ifdef COMMONWIN
	case 3: /* need to close the window we've created */
		ser_sr24_wincb_close(data);
#endif
	}

	data->event_handler.type = 0;

	switch(event->type)
	{
	case 0: /* disable event */
		data->event_handler.type = 0;
		result=0;
		break;

	case 1: /* callbacks */
		if( (result=ser_sr24_create_eventthread(data)) )
			break;
			
		data->event_handler.type = 1;
		data->event_handler.eventfn = event->eventfn;
		data->event_handler.extra = event->extra;
		result=0;
		break;

	case 2: /* Windows messages */
#ifdef LINUX
		result=SER_NOT_SUPPORTED;
#endif
#ifdef COMMONWIN


		data->event_handler.type = 2;
		result=0;
#endif
		break;

	case 3: /* Windows callback */
#ifdef LINUX
		result=SER_NOT_SUPPORTED;
#endif
#ifdef COMMONWIN
		if( (result=ser_sr24_wincb_create(data)) )
			break;
			
		data->event_handler.type = 3;
		data->event_handler.eventfn = event->eventfn;
		data->event_handler.extra = event->extra;
		result=0;
#endif
		break;
	default:
		result = SER_INVALID_EVENTTYPE;
	}

	ser_mutex_release(&(data->mutex));
	return(result);
}

int CALLCONVENTION ser_sr24_wait_event(ser_sr24_data* data, int* events, unsigned int length, unsigned int timeout, int* value)
{
/* This function waits for any of several events (in an array)
 * to occur. If an event occurs that is wanted, that event
 * id is returned by the function and the event's value is
 * written into the variable pointed to by value.  Since event id's 
 * cannot be negative, if a negative value is returned, it is an error
 * code.  If a value of 0 is passed for the length parameter,
 * the function will return any event that occurs.
 * A timeout value may be specified by the timeout parameter
 * This value is in milliseconds.  If no desired event occurs
 * within the specified amount of time, the function returns
 * with the timeout error.  If no timeout is desired, a value
 * of zero may be passed for timeout.
 */
	unsigned int *ptout;
	int x;
	int retval;
	int result=0;

	if(data == 0)
		return(SER_INVALID_HANDLE);
	
	if(timeout)
		ptout = &timeout;
	else
		ptout = 0;

	if(ser_event_prewait(&(data->synch_event)))
		return(SER_THREAD_ERROR);

	if(length)/* if there are events in the array */
	{
		while(result == 0)/* loop until we have something to return */
		{
			if( (retval=ser_event_wait(&(data->synch_event), ptout)) )
			{
				if(retval == 1)
					result = SER_TIMEOUT;
				else
					result = SER_THREAD_ERROR;
				break;
			}
			for(x=0; x<length; x++)
			{
				if(data->last_event == events[x])
				{
					result = data->last_event;
					if(value)
						*value = data->last_event_value;
					break; /* get out of for loop. the result value
					        * will end the while loop.
					        */
				}
			}
		}
	}
	else
	{ /* otherwise wait for any event to occur and return the first one */
		if( (retval=ser_event_wait(&(data->synch_event), ptout)) )
		{
			if(retval == 1)
				result = SER_TIMEOUT;
			else
				result = SER_THREAD_ERROR;
		}
		else /* success */
		{
			result = data->last_event;
			if(value)
				*value = data->last_event_value;
		}
	}
	
	ser_event_release(&(data->synch_event));
	return(result);
}

int CALLCONVENTION ser_sr24_close(ser_sr24_data* data)
{/* This function closes the device.  It returns 0 no matter
  * what at this point.
  * Regardless, the device will be closed, and the data
  * structure will be freed from allocated memory
  * 
  */
	/* don't need to check return values, because we want to close
	 * no matter what.
	 */
	ser_event_spec evspec;

	if(data == 0)
		return(SER_INVALID_HANDLE);

	evspec.type = 0;

	/* set event handler type to zero so anything such as an extra thread or window
	 * will be destroyed
	 */
	ser_sr24_set_event_handler(data, &evspec);

	/* The quitflags are protected by the predicate mutex of the event */
	ser_event_prewait(&(data->event));

	data->quitflag=1; //tell reading thread we are quitting.

	/* we'll send out a ping command so the reading thread
	 * will be awakened.  If the board doesn't ping back, then the reading
	 * thread will eventually wake due to serial timeouts.
	 */
	ser_serial_write(&(data->serial), "P\r", 2);

	
	/* now wait for thread to notify me that it has exited */

	while( data->quitflag ) /* wait until the thread sets quitflag 
	                         * back to zero */
	{
		ser_event_wait(&(data->event), 0);
	}

	/* IF we are here, then replies have come in, we still
	 * own the event.
	 */

	ser_serial_close(&(data->serial));
	
	ser_event_release(&(data->event));
	
	ser_event_destroy(&(data->event));
	ser_mutex_destroy(&(data->mutex));
	ser_event_destroy(&(data->synch_event));

#ifdef COMMONWIN
	/* If we're on a Windows machine, take this device out of the table
	 * of opened devices
	 */
	ser_sr24_table_remove(data);
#endif

	/* Finally, free the data structure we allocated previously */
	free(data);
	
	return(0);
}

char* CALLCONVENTION ser_sr24_error_string(int errorcode)
{
	static char buffer[300]; /* big enough for all error strings */

	switch(errorcode)
	{
	case SER_SUCCESS:
		strcpy(buffer, "SER_SUCCESS: No error occurred.");
		break;
	case SER_GENERAL_ERROR:
		strcpy(buffer, "SER_GENERAL_ERROR: A miscellaneous error has occurred.");
		break;
	case SER_SERIAL_ERROR:
		strcpy(buffer, "SER_SERIAL_ERROR: An error occurred either opening, reading, or writing to the serial port.");
		break;
	case SER_THREAD_ERROR:
		strcpy(buffer, "SER_THREAD_ERROR: Internal error occurred in the threads library.");
		break;
	case SER_NOT_SUPPORTED:
		strcpy(buffer, "SER_NOT_SUPPORTED: Device does not support this function.");
		break;
	case SER_PIPE_ERROR:
		strcpy(buffer, "SER_PIPE_ERROR: Internal error creating pipe.");
		break;
	case SER_READTHREAD_ERROR:
		strcpy(buffer, "SER_READTHREAD_ERROR: Internal error creating the reading thread.");
		break;
	case SER_EVENTTHREAD_ERROR:
		strcpy(buffer, "SER_EVENTTHREAD_ERROR: Internal error creating the event thread.");
		break;
	case SER_MALLOC_ERROR:
		strcpy(buffer, "SER_MALLOC_ERROR: Error allocating memory.");
		break;
	case SER_WINDOWS_ERROR:
		strcpy(buffer, "SER_WINDOWS_ERROR: Error creating a window.");
		break;
	case SER_INVALID_HANDLE:
		strcpy(buffer, "SER_INVALID_HANDLE: Handle specified is invalid.  No device is open with that handle.");
		break;
	case SER_INVALID_MODEL:
		strcpy(buffer, "SER_INVALID_MODEL: Invalid model specified.");
		break;
	case SER_INVALID_BAUD:
		strcpy(buffer, "SER_INVALID_BAUD: Invalid baud rate constant specified.");
		break;
	case SER_INVALID_PORT:
		strcpy(buffer, "SER_INVALID_PORT: The port specified does not exist.");
		break;
	case SER_INVALID_PORTMODE:
		strcpy(buffer, "SER_INVALID_PORTMODE: The mode specified is either unknown or unavaible on this port.");
		break;
	case SER_INVALID_PORTVALUE:
		strcpy(buffer, "SER_INVALID_PORTVALUE: Output value specified for port is out of the valid range.");
		break;
	case SER_INVALID_EVENTID:
		strcpy(buffer, "SER_INVALID_EVENTID: Event ID specified is out of valid range.");
		break;
	case SER_INVALID_EVENTTYPE:
		strcpy(buffer, "SER_INVALID_EVENTTYPE: Event type specified is invalid.");
		break;
	case SER_INVALID_BIT:
		strcpy(buffer, "SER_INVALID_BIT: Bit number specified is out of valid range.");
		break;
	case SER_INVALID_PROPERTY:
		strcpy(buffer, "SER_INVALID_PROPERTY: Property specified is not valid for this port.");
		break;
	case SER_INVALID_CHANNEL:
		strcpy(buffer, "SER_INVALID_CHANNEL: Analog channel number specified is invalid.");
		break;
	case SER_SERIALPORT_CONFLICT:
		strcpy(buffer, "SER_SERIALPORT_CONFLICT: Serial port is already in use.");
		break;
	case SER_DEVICE_NOTFOUND:
		strcpy(buffer, "SER_DEVICE_NOTFOUND: Device is not present on the serial port, or is not powered on.");
		break;
	case SER_WRONG_PORTMODE:
		strcpy(buffer, "SER_WRONG_PORTMODE: The action requested can't be performed while the port is in it's current mode.");
		break;
	case SER_EVENT_CONFLICT:
		strcpy(buffer, "SER_EVENT_CONFLICT: Bit events on a port cannot be registered when a byte event exists on that port.");
		break;
	case SER_TIMEOUT:
		strcpy(buffer, "SER_TIMEOUT: Timeout occurred.  This could be due to a communication problem with the serial board.");
		break;
	default:
		strcpy(buffer, "Unknown error code.");
		break;
	}
	return(buffer);
}

#ifdef COMMONWIN
/* On windows provide a helper function for Visual Basic
 * to get the address of a class object.
 */
int CALLCONVENTION ser_returnvalue(int value)
{
	return(value);
}

#endif