/* ser_sr24_helper.c
 * 
 * This file contains the internal helper functions used for
 * the SR24 device.  This file separates these helper functions
 * from the visible functions of the API, which are in ser_sr24.c
 * 
 * Copyright (C) 2000, Winford Engineering.
 * See file LICENSE.txt for details.
 * 
 * Winford Engineering
 * www.winfordeng.com
 */

#ifdef WIN95
#define COMMONWIN
#endif
#ifdef WINNT
#define COMMONWIN
#endif

#include "ser_sr24.h"
#include "ser_sr24_helper.h"
#include "ser_threads.h"
#include "ser_serial.h"
#include "ser_error.h"

#ifdef COMMONWIN
#include <stdio.h>
#include <math.h>
#include <windows.h>
#include <malloc.h>
#endif

#ifdef LINUX
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <malloc.h>
#endif

void ser_sr24_dispatch_event(ser_sr24_data* data, int id, int value)
{/* This function releases any threads waiting for synchronous
  * events, as well as using the selected event notification
  * mechanism to notify the user's application of the event
  */
	/* We use a event structure to pass event info to the third thread.
	 * This is not necessary in windows since we do the same thing by using 
	 * Windows messages
	 */
	ser_sr24_eventblock evb;
#ifdef COMMONWIN
	ser_sr24_eventblock *pevb; /* pointer to event block if we have to allocate one */
	DWORD nwritten;
#endif

	if( ser_event_prewait(&(data->synch_event)) == 0 )
	{/* continue if successful */
		data->last_event = id;
		data->last_event_value = value;
		ser_event_broadcast(&(data->synch_event));
		ser_event_release(&(data->synch_event));
	}
	
	switch(data->event_handler.type)
	{
	case 0:
		break;
	case 1: /* callback function */
		evb.id = id;
		evb.value = value;
		evb.extra = data->event_handler.extra;
#ifdef LINUX

		/* Now push the structure through the pipe for the event thread to see */
		write(data->pipe_event[1], &evb, sizeof(evb));
		/* The write call may fail if the pipe is full.  So be it.
		 * This will only happen if events are occurring way too fast,
		 * if the user's event procedure is taking way too long,
		 * or something else is terribly wrong
		 */

		/* We don't handle events this way anymore, because it ties up the reading thread.
		 * We now perform this call in the event #ifdef LINUX
		break;
#endif
thread (below).
		 * data->event_handler.eventfn(id, value);
		 */
#endif

#ifdef COMMONWIN
		WriteFile(data->pipe_event[1], &evb, sizeof(evb), &nwritten, NULL);
#endif

		break;
	case 2: /* Windows message */
#ifdef COMMONWIN
		PostMessage(data->event_handler.window, 
		            data->event_handler.msgid, 
		            id, /* wParam is the event ID */
		            value); /* lParam is the event value */
#endif		

		break;
	case 3: /* Special callback */
#ifdef COMMONWIN
		/* it's the responsibility of the window procedure to free this memory */
		pevb = (ser_sr24_eventblock*)malloc(sizeof(ser_sr24_eventblock));
		
		if(pevb == 0)
			break;

		pevb->id = id;
		pevb->value = value;
		pevb->extra = data->event_handler.extra;

		PostMessage(data->wincb, WM_USER, (WPARAM)pevb, (LPARAM)data);
#endif

		break;
	}
}


void ser_sr24_process_input(ser_sr24_data* data, char* input)
{
/* This function processes the data that is read from the serial
 * board.  This data includes events, result data from port reads, 
 * and so on.
 */
	int port;
	int bit;
	int value;
	char buf[10]; /* a little space to work with 
	               * At this point, the longest return string is for
	               * analog channels with format:
	               *  a.n=ddd\r 8 characters + 1 NULL = 9
	               */
	
	switch(*input)
	{
	case 'a': /* result of reading analog channel. */
		/* format: a.n=ddd */
		/* we'll use bit variable to store the channel number,
		 * just to save space
		 */
		if(strlen(input) < 7)break;

		/* get the channel number, store in bit */
		buf[0] = input[2];
		buf[1] = '\0';

		bit = strtol(buf, 0, 10);
		if((bit < SER_SR24_CHAN_MIN) || (bit > SER_SR24_CHAN_MAX))
			break;

		/* get the value of the channel */
		strncpy(buf, input + 4, 3); /* values are always 3 characters (padded). */
		buf[3] = '\0';
		value = strtol(buf, 0, 16); /* value is returned in hex */

		/* we have the data now. store and set the clean bits. */
		/* first lock the event before we modify stuff */
		if(ser_event_prewait(&(data->event)))break;

		data->analogvalue[bit] = value;
		data->analogclean |= (int)pow(2, bit);

		/* now before we release the event, signal that data has come in */
		ser_event_broadcast(&(data->event));
		ser_event_release(&(data->event));
		break;
		
	case 'P': /* result of Port Input. store the value */
		/* format: Pn=dd */
		if(strlen(input) < 5)break; /* must have at least 5 characters */

		/* first, get the port number */
		buf[0] = input[1];
		buf[1] = '\0';
		port = strtol(buf, 0, 10);
		if((port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX))
			break;
			
		/* now get the value */
		strncpy(buf, input + 3, 2);
		buf[2] = '\0';
		value = strtol(buf, 0, 16);/* value is returned in hexadecimal */

		/* now that we have the data, store it and set the clean bits */
		if(ser_event_prewait(&(data->event)))break;
		
		data->portvalue[port-1]=value;
		data->portclean |= (int)pow(2, port-1);

		ser_event_broadcast(&(data->event));
		ser_event_release(&(data->event));
		break;
		
	case 'p': /*result of bit input. store the value */
	    /* format pn.b=d */
		if(strlen(input) < 6)break; /* must have at least 6 characters */

		/* first, get the port number */
		buf[0] = input[1];
		buf[1] = '\0';
		port = strtol(buf, 0, 10);
		if((port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX))
			break;

		/* get the bit number */
		buf[0] = input[3]; /* terminated above */
		bit = strtol(buf, 0, 10);
		if((bit < SER_SR24_BIT_MIN) || (bit > SER_SR24_BIT_MAX))
			break;

		/* now get the value */
		buf[0] = input[5];
		value = strtol(buf, 0, 10);

		/* now that we have the data, store it and set the clean bits */
		if(ser_event_prewait(&(data->event)))break;

		if(value) /* do we want to turn the bit on or off?*/
		{
			data->bitvalue[port-1] |= (int)pow(2, bit);
		}
		else
		{
			data->bitvalue[port-1] &= ~(int)pow(2, bit);
		}

		data->bitclean[port-1] |= (int)pow(2, bit);

		ser_event_broadcast(&(data->event));
		ser_event_release(&(data->event));
		break;
		
	case 'B': /* port event has occured */
		if(strlen(input) < 5)
			break;

		/* first, get the port number */
		buf[0] = input[1];
		buf[1] = '\0';
		port = strtol(buf, 0, 10);
		if((port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX))
			break;
			
		/* now get the value */
		strncpy(buf, input + 3, 2);
		buf[2] = '\0';
		value = strtol(buf, 0, 16);/* value is returned in hexadecimal */

		if(data->portevent[port-1] == 0)/*if the event is not wanted, quit */
			break;

		ser_sr24_dispatch_event(data, data->portevent[port-1], value);
		break;
		
	case 'b': /* bit event has occured */
		if(strlen(input) < 6)break;

		/* get the port number */
		buf[0] = input[1];
		buf[1] = '\0';
		port = strtol(buf, 0, 10);
		if((port < SER_SR24_PORT_MIN) || (port > SER_SR24_PORT_MAX))
			break;

		/* now get the bit number */
		buf[0] = input[3];
		/*buf[1] is already NULL*/
		bit = strtol(buf, 0, 10);
		if((bit < SER_SR24_BIT_MIN) || (bit > SER_SR24_BIT_MAX))
			break;

		/* now get the value */
		buf[0] = input[5];
		value = strtol(buf, 0, 10);

		if(data->bitevent[port-1][bit] == 0) /* if the event is not wanted, quit */
			break;

		ser_sr24_dispatch_event(data, data->bitevent[port-1][bit], value);
		break;
		
	case 'G': /* result of a ping command */

		if(data->verifyflag)
		{
			ser_event_prewait(&(data->event));
			data->verifyflag = 0; /* show that we got the message */
			ser_event_broadcast(&(data->event));
			ser_event_release(&(data->event));
		}

		break;
	}
}


/* ******************************************************************
 * readthread function - main thread for processing device input 
 * ******************************************************************
 */
#ifdef LINUX
void* ser_sr24_readthread(void* arg)
#endif
#ifdef COMMONWIN
DWORD WINAPI ser_sr24_readthread(void* arg)
#endif
{
	/* This function is never actually "called" like a normal function is,
	 * but it is used as the thread function for reading data from the serial
	 * board.  In other words, the new thread executes this function, but I 
	 * never actually make a call to this function in my own code.
	 * 
	 * This function recieves a pointer to the ser_sr24_data
	 * structure as the arg parameter
	 */
	/* at this point, I think the longest input string will
	 * be a.n=xxx\r ( 8 + NULL ) = 9. To be safe, we'll allocate a bit more
	 */
#define BUFSIZE 20
	char buf[BUFSIZE];
	int pos; /* indicates how far we've filled our input buffer */
	int numread; /*receives the number of bytes we've read*/
	int retval;
	ser_sr24_data* data = (ser_sr24_data*)arg;

#ifdef LINUX	
	sigset_t mask;
	sigset_t oldmask;

	/* set signal handling mask to block all signals */
	sigfillset(&mask);
	pthread_sigmask(SIG_BLOCK, &mask, &oldmask);
#endif LINUX

#ifdef COMMONWIN
	/* On Windows, place a handle to this thread in the data
	 * structure.  This allows the thread to be terminated
	 * if the dll is unloaded dynamically.
	 */
	if(DuplicateHandle(GetCurrentProcess(), /* source process */
	                GetCurrentThread(),  /* handle to duplicate */
	                GetCurrentProcess(), /* target process */
	                &(data->readthread_handle), /* pointer to duplicate handle */
	                0, /* ignored because of SAME_ACCESS */
	                FALSE, /* not inheritable */
	                DUPLICATE_SAME_ACCESS) == 0)
	{
		/* If we failed to duplicate the handle, set the
		 * handle in the data structure to invalid to indicate this.
		 */
		data->readthread_handle = INVALID_HANDLE_VALUE;
	}
	
#endif
	
	while( data->quitflag == 0 )/* master loop to keep everything going */
	{
		pos=0;
		while( (retval=ser_serial_readline(&(data->serial), buf+pos, BUFSIZE-pos, &numread)) )/* loop til we get a complete line*/
		{/* NOTE: we  may want to do something here sometime in case the readline fails */
			if(data->quitflag)
				break;
			pos += numread;
		}

		if(data->quitflag)
			break;

		/* if we are here, we have a line in buf that should be processed */
		ser_sr24_process_input(data, buf);
	}

	/* if we're here, then quitflag has been set */
	ser_event_prewait(&(data->event));
	data->quitflag = 0; /* show that we got the message */
	ser_event_broadcast(&(data->event));
	ser_event_release(&(data->event));
	
	return(0);
}

/* ******************************************************************
 * eventthread function - thread for processing events, calling callbacks.
 * ******************************************************************
 */
#ifdef LINUX
void* ser_sr24_eventthread(void* arg)
#endif
#ifdef COMMONWIN
DWORD WINAPI ser_sr24_eventthread(void* arg)
#endif
{
	/* This function is similar to the readthread function in that it is
	 * only executed by a new thread.
	 * 
	 * The event thread is dedicated to executing the callback events.
	 * Whenever an event occurs and the reading thread determines the 
	 * application should be notified through a callback, the information
	 * is passed to this thread through a pipe, and this thread does the
	 * actual calling of the user's callback function.
	 */
	 
	/* this function recieves a pointer to the ser_sr24_data
	 * structure as the arg parameter
	 */
	/* When this thread is created, the pipe should have already been created,
	 * since this thread immediately begins reading the pipe for events.
	 */

	int have=0;
	ser_sr24_eventblock evb; /* event information which will be read from the pipe */
	ser_sr24_data* data = (ser_sr24_data*)arg;

#ifdef COMMONWIN
	DWORD ret=0;

	/* On Windows, place a handle to this thread in the data
	 * structure.  This allows the thread to be terminated
	 * if the dll is unloaded dynamically.
	 */
	if(DuplicateHandle(GetCurrentProcess(), /* source process */
	                GetCurrentThread(),  /* handle to duplicate */
	                GetCurrentProcess(), /* target process */
	                &(data->evthread_handle), /* pointer to duplicate handle */
	                0, /* ignored because of SAME_ACCESS */
	                FALSE, /* not inheritable */
	                DUPLICATE_SAME_ACCESS) == 0)
	{
		/* If we failed to duplicate the handle, set the
		 * handle in the data structure to invalid to indicate this.
		 */
		data->evthread_handle = INVALID_HANDLE_VALUE;
	}
	
#endif

#ifdef LINUX
	int ret;
	sigset_t mask;
	sigset_t oldmask;

	/* set signal handling mask to block all signals for this thread*/
	sigfillset(&mask);
	/* Don't block SIGCHLD */
	sigdelset(&mask, SIGCHLD);
	
	pthread_sigmask(SIG_SETMASK, &mask, &oldmask);



	/* loop until a full event block has come in.  Then loop again infinitely until pipe is closed 
	 * Loop statements are different between operating systems because of different read call
	 */
	while( (ret=read(data->pipe_event[0], &evb + have, sizeof(evb) - have)) > 0 )
#endif /* LINUX */
#ifdef COMMONWIN
	while( ReadFile(data->pipe_event[0], &evb + have, sizeof(evb) - have, &ret, NULL))
#endif
	{
		have += ret;
		if(have == sizeof(evb))
		{
			data->event_handler.eventfn(evb.id, evb.value, evb.extra);
			have = 0;
		}
	}

	/* If we're here, the pipe's been closed or there was an error.
	 * All we can do is close our end of the pipe and exit the thread.
	 */

#ifdef LINUX	 
	close(data->pipe_event[0]);
#endif
#ifdef COMMONWIN
	CloseHandle(data->pipe_event[0]);
#endif

	/* pipe has been closed, set flag back to zero to notify that we're exiting */
	ser_event_prewait(&(data->event));
	data->event_quitflag = 0; /* show that we got the message */
	ser_event_broadcast(&(data->event));
	ser_event_release(&(data->event));

	return(0);
}

int ser_sr24_create_eventthread(ser_sr24_data *data)
{
	/* This function creates the event handler thread for use in callback
	 * event notification.  This function assumes that no event thread
	 * has already been created.
	 * 
	 * This function returns 0 on success or a negative error code
	 * on failure.
	 */

	int flags; /* file descriptor flags */
	int erresult;
	
#ifdef LINUX
	/* Open a pipe for communication between the reading thread and the event thread */
	if( pipe(data->pipe_event) )
	{
		erresult = SER_PIPE_ERROR;
		goto error_1;
	}

	/* Put the writing end of the pipe in nonblocking mode, so the reading thread
	 * doesn't hang while trying to write a new event to the pipe.
	 * Note that if the pipe is not full, then the pipe will behave like normal
	 */
	flags = fcntl(data->pipe_event[1], F_GETFL);
	if(fcntl(data->pipe_event[1], F_SETFL, flags | O_NONBLOCK) == -1)
	{
		erresult = SER_PIPE_ERROR;
		goto error_2;
	}

	
#endif

#ifdef COMMONWIN

	if(CreatePipe(&(data->pipe_event[0]), &(data->pipe_event[1]), NULL, PIPE_BUFFER) == 0)
	{
		erresult = SER_PIPE_ERROR;
		goto error_1;
	}
	
#endif 


	if( ser_thread_create(ser_sr24_eventthread, (void*)data) )
	{
		erresult=SER_EVENTTHREAD_ERROR;
		goto error_2;
	}

	return(0);
error_2: 
/* We need to close the pipe fd's. */
#ifdef LINUX
	close(data->pipe_event[0]);
	close(data->pipe_event[1]);
#endif /* LINUX */

#ifdef COMMONWIN
	CloseHandle(data->pipe_event[0]);
	CloseHandle(data->pipe_event[1]);
#endif /* COMMONWIN */

error_1:
	return(erresult);
}

void ser_sr24_close_eventthread(ser_sr24_data* data, int mode)
{
	/* This function closes the event thread for the API
	 * 
	 * This function should only be called when the event thread is actually
	 * present, otherwise it will hang waiting for a response back.
	 * 
	 * The exception is that if mode is nonzero, no waiting will be done
	 * for the thread.  This function will only attempt to close it, but
	 * will not wait for confirmation.
	 * 
	 * This function attempts to close the event thread regardless of
	 * conditions, so it does not return a value.
	 */

	/* obtain the predicate mutex, which protects the quitflag */
	
	ser_event_prewait(&(data->event));

	data->event_quitflag=1; /* set the flag for the event thread.  
	                         * the event thread will set this back to zero when it exits
	                         */

	/* close the pipe, causing the event thread to exit */
#ifdef LINUX
	close(data->pipe_event[1]); /* it should only be necessary to close the writing
	                                     * end of this pipe.  The read call on the other end
	                                     * will fail, and the event thread will exit.
	                                     */
#endif

#ifdef COMMONWIN
	CloseHandle(data->pipe_event[1]); /* reading end should close itself */
#endif

	if( mode == 0 ) /* should be default behaviour */
	{
		while( data->event_quitflag ) /* wait until the event thread has quit */
		{
			ser_event_wait(&(data->event), 0);
		}
	}

	ser_event_release(&(data->event));
}

void ser_sr24_reset_board_data(ser_sr24_data *data)
{
	/* This function resets the items in data
	 * that reflect the state of the serial board.
	 * 
	 * This function should be called during the opening
	 * of a device, and also during a reset_board call.
	 * 
	 * The data structure is not protected by a mutex
	 * or anything in this function.  If that is desired,
	 * it should be done before the function call is made.
	 */

	data->mode[0] = 0;
	data->mode[1] = 0;
	data->mode[2] = 0;
	data->portevent[0] = 0;
	data->portevent[1] = 0;
	data->portevent[2] = 0;
	memset(data->bitevent, 0, sizeof(data->bitevent));
}


#ifdef COMMONWIN

int ser_sr24_wincb_create(ser_sr24_data *data)
{
	/* This function creates the window necessary for implementing event
	 * handler type 3 of the API.  If everything is created successfully,
	 * this function returns 0, otherwise an appropriate negative error
	 * code
	 */
	HANDLE win;
	WNDCLASS wc;
	int erresult=0;

	wc.style = 0;
	wc.lpfnWndProc = ser_sr24_wincb_proc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = 0;
	wc.hIcon = 0;
	wc.hCursor = 0;
	wc.hbrBackground = 0;
	wc.lpszMenuName = 0;
	wc.lpszClassName = "ser_sr24_wincb_class";

	if(RegisterClass(&wc) == 0)
	{
		erresult = SER_WINDOWS_ERROR;
		goto error_1;
	}
		
	if( (win=CreateWindow("ser_sr24_wincb_class", /* class name */
	                      "ser_sr24_wincb", /* window name */
	                      0, /* window styles */
	                      0, 0, /* x, y */
	                      50, 50, /* width, height */
	                      0, /* parent window */
	                      0, /* menu */
	                      0, /* instance */
	                      0 /* param */
	                      )) == 0 )
	{
		erresult = SER_WINDOWS_ERROR;
		goto error_2;
	}

	/* If we're here, we're successful.  Just store the handle in the
	 * data structure and exit
	 */
	data->wincb = win;
	
	return(0);
error_2: /* failed to create window.  need to get rid of class */
	UnregisterClass("ser_sr24_wincb_class", 0);
	
error_1: /* no cleanup necessary */
	return(erresult);
}

int ser_sr24_wincb_close(ser_sr24_data *data)
{
	/* This function gets rid of the window and class used for this method
	 * of event notification.
	 * 
	 * Assumes the stuff has been created.  No checks made.
	 * 
	 * Returns 0 if there were no problems, nonzero if there were problems.
	 * Everything possible is closed regardless.
	 */
	int fail;
	
	fail = !(DestroyWindow(data->wincb));
	fail |= !(UnregisterClass("ser_sr24_wincb_class", 0));

	return(fail);
}

long WINAPI ser_sr24_wincb_proc(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	/* Message handling function for the window we've created.
	 * We'll use the WM_USER message for the events for the serial board
	 * 
	 * wParam contains a pointer to an eventblock structure that has 
	 * been malloced, and contains information about the event that occured.
	 * We must free this structure.
	 * 
	 * lParam contains a pointer to the entire device data structure.
	 * We definitely must NOT free this structure.
	 */
	ser_sr24_eventblock *pevb;
	ser_sr24_data *data;
	
	switch(Message)
	{
	case WM_USER:
		pevb = (ser_sr24_eventblock*)wParam;
		data = (ser_sr24_data*)lParam;

		data->event_handler.eventfn(pevb->id, pevb->value, pevb->extra);
		
		free(pevb); /* free the allocated memory */
		
		break;
	}

	
	return(DefWindowProc(hWnd, Message, wParam, lParam));
}

/* On windows, we'll keep a list of opened devices, up to TABLELEN long,
 * and when a device is opened, the address of the structure will be put
 * into the table.  When a device is closed, it's entry will be removed
 * from the table.
 * 
 * When the Dll unloads, all unclosed devices will be closed.
 * 
 * This is primarily for Visual Basic, where in development mode, the Dll
 * may be loaded and unloaded without the process ending.  Because of this,
 * it's important to totally clean up during unload.
 * 
 * Note that if the table runs out of room, it won't prohibit more devices
 * from being opened, they just won't be automatically unloaded.
 * 
 * NOTE: If an entry contains a zero, it is considered empty.
 */


static void ser_sr24_terminate(ser_sr24_data *data)
{ 
/* When unloading, we can't just use the close function, because during DLL
 * unload, the other threads seem to be in a disabled state, so the close
 * function hangs around indefinitely waiting for response from the other
 * threads, and never gets it.
 * 
 * We'll use a special terminate function to take the place of a close
 * function in these special situations.  This is only necessary
 * on windows.
 */

	/* If we have anything created for event handling, get rid of that
	 * now.  This could be an extra window, or a thread.
	 */

	switch(data->event_handler.type)
	{
	case 1: /* need to kill event thread */
		if(data->evthread_handle != INVALID_HANDLE_VALUE) /* if valid */
			TerminateThread(data->evthread_handle, 0);

		CloseHandle(data->pipe_event[1]); /* close pipe handles */
		CloseHandle(data->pipe_event[0]);

		break;
	/* no cleanup necessary for plain Windows messages */
	case 3:
		/* it's safe to call cleanup function to destroy the window */
		ser_sr24_wincb_close(data);
		break;
	}

	/* Terminate reading thread if possible */
	if(data->readthread_handle != INVALID_HANDLE_VALUE)
		TerminateThread(data->readthread_handle, 0);

	/* close serial connection and synchronization objects */
	ser_serial_close(&(data->serial));
	
	ser_event_destroy(&(data->event));
	ser_mutex_destroy(&(data->mutex));
	ser_event_destroy(&(data->synch_event));

	/* Take this device out of the table
	 * of opened devices, even though it's not necessary.
	 */
	ser_sr24_table_remove(data);

	/* Finally, free the data structure we allocated previously */
	free(data);
}

#define TABLELEN 32
static ser_sr24_data *device_table[TABLELEN]; /* table of pointers */

static void ser_sr24_table_init()
{
	int i;

	for(i=0; i<TABLELEN; i++)
		device_table[i] = 0;
}

void ser_sr24_table_add(ser_sr24_data *data)
{ /* This function adds the specified address to the device table */
	int i;

	for(i=0; i<TABLELEN; i++)
	{
		if(device_table[i] == 0)
		{
			device_table[i] = data;
			break;
		}
	}
}

void ser_sr24_table_remove(ser_sr24_data *data)
{ /* This function removes the specified address from the device table */
	int i;

	for(i=0; i<TABLELEN; i++)
	{
		if(device_table[i] == data)
		{
			device_table[i] = 0;
			break;
		}
	}
}

static void ser_sr24_table_cleanup()
{
	int i;
	ser_sr24_data *data;

	for(i=0; i<TABLELEN; i++)
	{
		if(device_table[i])
		{
			ser_sr24_terminate(device_table[i]);
		}
	}
}

BOOL WINAPI __declspec(dllexport) LibMain(HINSTANCE hDllInst, DWORD fdwReason, LPVOID lpvReserved)
{
	/* This function is called whenever this Dll is loaded or unloaded
	 * by a process. 
	 * 
	 * NOTE: When unloading, lpvReserved is NULL if the Dll is being
	 * dynamically unloaded with FreeLibrary.  In this case, we'll
	 * clean up our devices, otherwise we won't.
	 */
	int i;

	switch(fdwReason)
	{
	case DLL_PROCESS_ATTACH:
		ser_sr24_table_init();
			
		break;
	case DLL_PROCESS_DETACH:
		if(lpvReserved == NULL) /* if we're being unloaded by FreeLibrary */
			ser_sr24_table_cleanup();

		/* otherwise, the process is ending, and we'll just let everything
		 * die by itself
		 */
		break;
	}
	return(TRUE);
}

#endif // COMMONWIN