/* 
 * This is a sample application for the serial I/O board 
 * discussed in Winford Engineering's article in Nuts & 
 * Volts magazine.
 * 
 * This sample illustrates how to create and use a simple 
 * callback function for event notification.  It also uses 
 * the analog capabilities of board.
 * 
 * SETUP REQUIREMENTS:
 * This sample uses the same circuitry setup as the Visual 
 * Basic sample that was briefly discussed in the magazine 
 * article.  Please refer to the schematic and explanation 
 * in the article for details on what to hook up to your 
 * serial I/O board.
 * 
 * OVERVIEW:
 * This sample continuously polls one analog channel of the 
 * serial board, which should have a potentiometer attached.
 * At the same time, the application has registered and 
 * awaits one hardware bit event, which will be triggered 
 * whenever the pushbutton is pressed or released.  The 
 * application maintains a count of the number of times the 
 * event has occurred.  The screen is updated with this 
 * information and the analog value at the rate of the 
 * main polling loop.
 */
#include <windows.h>
#include <stdio.h>
#include "ser_sr24.h"

#define SERIALPORT "COM1"
#define BAUD SER_9600


/* Number of milliseconds between each poll of analog value.
 * This also affects the updating of the screen.
 */
#define POLL 20

/* function to update the screen with new info */
void update_screen(int events, int analog);

/* define a type of data structure that will be shared 
 * between the callback and the main program.  The main 
 * program will declare one instance of this structure, 
 * and pass the address to the callback so it can use it 
 * too.
 */
typedef struct
{
	ser_sr24 handle; /* handle to the I/O board */
	int count;  /* number of events that have occurred */
} data_struct;
	

void __stdcall event_callback(int id, int value, void *extra)
{
	/* This function will be executed whenever the pushbutton 
	 * is pressed or released.  In this particular example, 
	 * we don't care about the new value of the event, even 
	 * though it tells us whether the button was pressed or 
	 * released.  We only care about the number of times the 
	 * event has occurred.
	 * 
	 * In the main function below, we set up the "extra" 
	 * parameter to point to the shared data structure,
	 * defined above, which holds the handle and the count 
	 * of times the event has occurred.  We will increment 
	 * the count in here.  We will also write the new 
	 * value out to port 3, so the LEDs will display the 
	 * value in binary form.`
	 */
	data_struct *data=(data_struct*)extra;

	/* Only count the event if it has an ID of 200
	 * (which we assigned to the pushbutton event in the 
	 * main program.  In this application, it's the only 
	 * event, so we wouldn't NEED to make this check, but 
	 * it illustrates the point.
	 */
	if(id == 200) /* pushbutton event */
	{
		/* First, increment the count.  Also, wrap the count 
		 * around if it is over 255, because nothing more than 
		 * 255 can be output to the 8-bit I/O ports.
		 */
	
		data->count = (data->count + 1) % 256;
	
		/* Now, output the new value to port 3 */
		ser_sr24_output_byte(data->handle, 3, data->count);
	
		/* That's all.  The main program will update the value 
		 * displayed on the screen during the next run of its
		 * polling loop.
		 */
	}
}
	
int main(int argc, char *argv[])
{
	ser_sr24 handle;
	ser_event_spec evspec;
	int result;
	int analog; /* the latest analog value */
	data_struct data;

	/* Open the board, checking for errors.   */
	if( (handle=ser_sr24_open(SERIALPORT, BAUD, &result)) == 0)
	{
		printf("Failed to open serial board: %s\n", ser_sr24_error_string(result));
		exit(1);
	}

	/* Initialize our shared data structure */
	data.handle = handle;
	data.count=0;
	
	/* Put Port 1 into analog mode for the potentiometer */
	result = ser_sr24_set_portmode(handle, 1, SR24_TYPE, SR24_TYPE_ANALOG);
	if(result)
		printf("Error setting analog mode: %s\n", ser_sr24_error_string(result));

	/* Set up the event-handling mechanism that we will use (callback).
	 * Also, set up the "extra" pointer to be the address of the data 
	 * structure.  That way, we can share the data structure with the 
	 * callback
	 */
	evspec.type = 1;
	evspec.eventfn = event_callback; /* the function defined above */
	evspec.extra = &data;

	if( (result=ser_sr24_set_event_handler(handle, &evspec)) )
		printf("Error setting event handler: %s\n", ser_sr24_error_string(result));
	
	/* Now, register the hardware event for the pushbutton.
	 * This is a bit event on Port 2, bit 0.  We'll give 
	 * it an ID of 200 
	 */
	if( (result=ser_sr24_register_event(handle, 2, 0, 200)) )
		printf("Error registering event: %s\n", ser_sr24_error_string(result));

	/* Now enter an infinite polling loop to monitor the 
	 * analog value, and update the screen
	 */
	while(1)
	{
		/* Get value of analog channel 0 */
		if(ser_sr24_input_analog(handle, 0, &analog))
			analog=0;

		update_screen(data.count, analog);

		/* Now sleep for a little bit so we don't 
		 * hog all the CPU time.
		 */
		Sleep(POLL);
	}

	/* If we weren't in an infinite loop, we would want 
	 * to close the board when we were done.
	 */
	ser_sr24_close(handle);
	return(0);
}

void update_screen(int events, int analog)
{ /* Function to update the screen with the latest information 
   * on analog value and event counts.  It displays a little 
   * text bar graph for the analog value and just displays the 
   * number of events that have occurred 
   */
#define GRAPH_WIDTH 50
	static char graph[GRAPH_WIDTH + 1];
	int on; /* number of characters that should be "on" */
	        /* number of "off" will be GRAPH_WIDTH - on */


	/* Calculate "on" characters */
	on = (analog * GRAPH_WIDTH) / (1023);

	/* Fill "on" characters with pound signs */
	memset(graph, '#', on);

	/* Fill the rest with dashes */
	memset(graph + on, '-', GRAPH_WIDTH - on);

	/* NULL terminate the string */
	graph[GRAPH_WIDTH] = 0;

	
	printf("Analog: %s (%-4d)  Events: %-3d\r", graph, analog, events);
	fflush(stdout);
}